/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Malte Kraus.
 *
 * Contributor(s):
 *  Malte Kraus <firefox@maltekraus.de> (Original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function adblockFilterExporter() {
  var This = this;
  window.addEventListener("load", function() { This._init(); }, false);
  window.addEventListener("unload", function() { This._uninit(); }, false);

  this.abp = this._service.abp;
}
adblockFilterExporter.prototype = {
  _service: Components.classes["@adblockfilters.mozdev.org/service;1"]
                      .getService().wrappedJSObject,
  abp: null,
  _up: null,

  uploadData: function adblockFilters__uploadData() {
    this._service.blackWhiteList = this.blackWhiteList.concat([]);
    this._service.doUpload();
  },
  _uninit: function adblockFilters__uninit() {
    Components.classes["@mozilla.org/observer-service;1"]
              .getService(Components.interfaces.nsIObserverService)
              .removeObserver(this, "adblockfilters-sync-finished");
  },
  observe: function adblockFilters__observe(subject, topic, errCode) {
    if(topic != "adblockfilters-sync-finished")
      return;
    if(parseInt(errCode) == Components.results.NS_OK) {
      this.blackWhiteList = this._service.blackWhiteList.concat([]);
      treeView.boxObject.invalidate();
    } else {
      document.documentElement.getButton("extra1").removeAttribute("disabled");
    }
  },
  _init: function adblockFilters___init() {
    let abp = this.abp;

    Components.classes["@mozilla.org/observer-service;1"]
              .getService(Components.interfaces.nsIObserverService)
              .addObserver(this, "adblockfilters-sync-finished", false);

    gAdblockFilterExport.blackWhiteList = gAdblockFilterExport._service.blackWhiteList.concat([]);

    function replaceMethod(obj, method, replacementMethod, after) {
      let orig = obj[method];
      obj[method] = function() {
        if(!after) orig.apply(this, arguments);
        replacementMethod.apply(this, let(a=Array.slice(arguments,0)) (a.splice(0,0,orig), a));
        if(after)  orig.apply(this, arguments);
      };
    }

    // when a blacklisted filter is modified, it should remain blacklisted
    replaceMethod(treeView, "stopEditor", function(origStopEditor, save, blur) {
      if(this.editedRow < 0 || !save)
        return;
      let info = this.getRowInfo(this.editedRow);
      if(!info[1] || typeof info[1] == "string" || !("text" in info[1]))
        return;
      let beforeFilter = info[1].text;
      if(!beforeFilter)
        return;
      let afterFilter = gAdblockFilterExport.abp.Filter.normalize(this.editor.value);
      if(afterFilter && afterFilter != beforeFilter) {
        let filterIndex = gAdblockFilterExport.blackWhiteList.indexOf(beforeFilter);
        if(filterIndex != -1) {
          gAdblockFilterExport.blackWhiteList.splice(filterIndex, 1, afterFilter);
        }
      }
    }, true);

    // remove blacklists when a filter is removed
    replaceMethod(treeView, "removeFilter", function(origRemove, group, pattern) {
      let special = group instanceof abp.SpecialSubscription, idx;
      if(pattern && typeof pattern != "string" && group && special &&
         (idx = gAdblockFilterExport.blackWhiteList.indexOf(pattern.text)) != -1)
        gAdblockFilterExport.blackWhiteList.splice(idx, 1);
    });

    // on apply, sync the blacklist changes with the service
    replaceMethod(window, "applyChanges", function() {
      gAdblockFilterExport._service.blackWhiteList = gAdblockFilterExport.blackWhiteList;
    });

    //
    replaceMethod(window, "onChange", function() {
      document.documentElement.getButton("extra1").removeAttribute("disabled");
    });


    // highlight personal filters
    replaceMethod(treeView, "getRowProperties", function(orig, row, properties) {
      let filter = this.getRowInfo(row)[1];
      if(!filter || !(filter instanceof abp.Filter))
        return;

      let blacklisted = gAdblockFilterExport.blackWhiteList.indexOf(filter.text) !== -1;
      properties.AppendElement(this.atoms["upload-blacklist-" + blacklisted]);
    });
    let (atomService = Components.classes["@mozilla.org/atom-service;1"]
                            .getService(Components.interfaces.nsIAtomService)) {
      treeView.atoms["upload-blacklist-true"] = atomService.getAtom("upload-blacklist-true");
      treeView.atoms["upload-blacklist-false"] = atomService.getAtom("upload-blacklist-false");
    };

    document.getElementById("listitem-context").addEventListener("popupshowing", function(e) {
      gAdblockFilterExport.onContextShowing(e);
    }, false);

    // add the upload button
    gAdblockFilterExport._addUploadButton();
  },
  _addUploadButton: function() {
    document.documentElement.buttons += ",extra1";
    let btn = document.documentElement.getButton("extra1");
    btn.hidden = false;

    let (strings = Components.classes["@mozilla.org/intl/stringbundle;1"]
                    .getService(Components.interfaces.nsIStringBundleService)
      .createBundle("chrome://adblockfilters/locale/abp-settings.properties")) {
      if(nsPreferences.getBoolPref("extensions.adblockfilters.sync.enabled"))
        btn.setAttribute("label", strings.GetStringFromName("sync-button"));
      else
        btn.setAttribute("label", strings.GetStringFromName("upload-button"));
    }

    btn.addEventListener("command", function(event) {
      event.stopPropagation();
      if(!E("applyButton").hasAttribute("disabled"))
        applyChanges();
      this.setAttribute("disabled", true);
      gAdblockFilterExport.uploadData();
    }, false);
  },

  blackWhiteList: null,

  showPreferences: function adblockFilters__showPreferences() {
    let wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
                       .getService(Components.interfaces.nsIWindowMediator);
    let win = wm.getMostRecentWindow("Adblockfilters:Settings");
    if(win) {
      win.focus()
    } else {
      window.open("chrome://adblockfilters/content/settings.xul", "_blank",
                  "chrome,titlebar,toolbar,centerscreen,modal=0,dialog=0");
    }
  },

  selectedFilters: function() {
    let abp = this.abp;
    return treeView.getSelectedInfo(false).filter(function(sel) {
      return sel[0] instanceof abp.SpecialSubscription &&
             sel[1] instanceof abp.Filter;
    }).map(function(sel) { return sel[1]; });
  },
  onContextShowing: function(event) {
    let selected = this.selectedFilters();
    let item = document.getElementById("context-toggle-uploader-blacklist");
    item.hidden = !selected.length;
    if(selected.length) {
      let notOnList = selected.some(function(sel) {
        return (this.blackWhiteList.indexOf(sel.text) == -1);
      }, this);
      item.setAttribute("checked", !!(this._service.useWhitelist ^ notOnList));
    }
  },
  toggleBlacklist: function() {
    let sel = this.selectedFilters();
    let onList = sel.some(function(selected) {
      return (this.blackWhiteList.indexOf(selected.text) == -1);
    }, this);
    sel.forEach(function(selected) {
      let idx = this.blackWhiteList.indexOf(selected.text);
      if(idx == -1 && onList)
        this.blackWhiteList.push(selected.text);
      else if(idx != -1 && !onList)
        this.blackWhiteList.splice(idx, 1);
    }, this);
  },

  QueryInterface: function(aIID) {
    if(aIID.equals(Ci.nsISupports) || aIID.equals(Ci.nsIObserver))
      return this;
    throw Componrnts.results.NS_ERROR_NO_INTERFACE;
  }
}
var gAdblockFilterExport = new adblockFilterExporter();