/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Malte Kraus.
 *
 * Contributor(s):
 *  Malte Kraus <firefox@maltekraus.de> (Original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function FilterCollector() {
}
FilterCollector.prototype = {
  /* blacklist */
  _blacklist: null,
  get service() {
    return Cc["@adblockfilters.mozdev.org/service;1"].getService().wrappedJSObject;
  },
  get useWhitelist() {
    return this.prefs.getBoolPref("extensions.adblockfilters.whitelist");
  },
  get blackWhiteList() {
    if(!this._blacklist) {
      let blacklist = [];
      let filters = this.getUserPatterns();
      function filterExists(filter1) {
        return filter1 && filters.some(function(filter2) { return filter1 == filter2.text; });
      }

      var prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService)
                    .getBranch("extensions.adblockfilters.blacklist.");
      var list = prefs.getChildList("", {});
      list.sort(function(a, b) { return a - b; });
      for(let i = 0; i < list.length; i++) {
        if(prefs.prefHasUserValue(list[i])) {
          let filter = prefs.getComplexValue(list[i], Ci.nsISupportsString).data;
          if(filterExists(filter))
            blacklist.push(filter);
          else
            prefs.clearUserPref(list[i]);
        }
      }
      this._blacklist = blacklist;
    }
    return this._blacklist;
  },
  set blackWhiteList(blacklist) {
    blacklist = blacklist.filter(function(f) { return f; });

    let equal = this.blackWhiteList.length == blacklist.length &&
                this._blacklist.every(function(filter, i) { return blacklist[i] == filter; });
    if(equal) return blacklist;

    let prefs = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService)
                  .getBranch("extensions.adblockfilters.blacklist.");
    let list = prefs.getChildList("", {});
    for(let i = 0; i < list.length; i++) {
      if(prefs.prefHasUserValue(list[i]) && !blacklist[list[i]])
        prefs.clearUserPref(list[i]);
    }

    let str = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
    blacklist.forEach(function(data, i) {
        str.data = data;
        prefs.setComplexValue(i + 1, Ci.nsISupportsString, str);
    });

    return this._blacklist = blacklist;
  },

  sortCallback: function(filter1, filter2) {
    if(filter1.type == "comment" && filter2.type == "comment")
      return 0;
    if(filter1.text < filter2.text)
      return -1;
    if (filter1.text > filter2.text)
      return 1;
    return 0;
  },
  /* collect and assemble filters from abp */
  getUserPatterns: function() {
    let abp = this.service.abp;
    let patterns = [];
    let doSort = nsPreferences.getBoolPref("extensions.adblockfilters.sortOutput", false);
    for each(let subscription in abp.FilterStorage.knownSubscriptions) {
      if(subscription instanceof abp.SpecialSubscription) {
        let subPatterns = subscription.filters.map(function(filter) {
          return {
            text: filter.text,
            disabled: (filter.disabled || subscription.disabled),
            type: (filter instanceof abp.CommentFilter) ? "comment" :
                  (filter instanceof abp.BlockingFilter || filter instanceof abp.RegExpFilter) ? "filterlist" :
                  (filter instanceof abp.WhitelistFilter) ? "whitelist" :
                  (filter instanceof abp.ElemHideFilter) ? "elemhide" : null,
            thirdParty: filter.thirdParty,
            collapse: filter.collapse,
            includeDomains: filter.includeDomains,
            excludeDomains: filter.excludeDomains
          };
        }, this);
        if(doSort)
          subPatterns = subPatterns.sort(this.sortCallback)
        patterns = patterns.concat(subPatterns);
      }
    }
    return patterns;
  },
  getData: function(patterns, blacklist) {
    let abp = this.service.abp;
    let versionComparator = abp.Utils.versionComparator;
    function max(a, b) {
      return (versionComparator.compare(a, b) > 0) ? a : b;
    }
    const MAX_VERSION = "1.2";
    function filterRequiresAbp(pattern) {
      let filter = pattern.text, type = pattern.type;
      if(type == "comment" || type == "invalid")
        return "0";
      if(filter.contentType & abp.RegExpFilter.typeMap.ELEMHIDE)
        return "1.2";
      if(filter.substr(0, 2) == "||" || filter.substr(0, 4) == "@@||" ||
         (type == "filterlist" && filter.indexOf("^") != -1) ||
         (type == "elemhide" && filter.indexOf("~") != -1 &&
          filter.indexOf("~") < filter.indexOf("#")))
        return "1.1";
      if(pattern.includeDomains != null || pattern.excludeDomains != null)
        return "1.0.1";
      if(pattern.thirdParty != null)
        return "1.0";
      if(pattern.collapse != null)
        return "0.7.5";
      if(abp.Filter.optionsRegExp.test(filter))
        return "0.7.1";
      if(type == "elemhide")
        return (/^#([\w\-]+|\*)(?:\(([\w\-]+)\))?$/.test(filter) ? "0.6.1" : "0.7");
      if(/^(?:@@)\|?/.test(filter) || filter.substr(-1) == "|" || type == "whitelist")
        return "0.6.1.2";
      return "0";
    }

    let data = [];

    const COMMENT_PREFIX = "extensions.adblockfilters.comment.";
    var stringBundle = Cc["@mozilla.org/intl/stringbundle;1"]
                         .getService(Ci.nsIStringBundleService)
              .createBundle("chrome://adblockfilters/locale/upload.properties");
    if(this.prefs.getBoolPref(COMMENT_PREFIX + "expire")) {
      let time = this.prefs.getIntPref(COMMENT_PREFIX + "expire.value");
      data.push(stringBundle.formatStringFromName("comment-expire", [time], 1));
    }
    if(this.prefs.getBoolPref(COMMENT_PREFIX + "redirect")) {
      let url = this.prefs.copyUnicharPref(COMMENT_PREFIX + "redirect.value");
      data.push(stringBundle.formatStringFromName("comment-redirect", [url], 1));
    }

    if(!patterns) patterns = this.getUserPatterns();
    if(!blacklist) blacklist = this.blackWhiteList;
    var minVersion = "0";
    Array.forEach(patterns, function(pattern) {
      if((blacklist.indexOf(pattern.text) != -1) ^ this.useWhitelist)
        return;
      if(pattern.disabled && pattern.type != "comment")
        return; //pattern.text = "!" + pattern.text;
      else if(minVersion != MAX_VERSION)
        minVersion = max(filterRequiresAbp(pattern), minVersion);
      data.push(pattern.text);
    });


    if(max(minVersion, "0.7.1pre") == minVersion)
      data.splice(0, 0, "[Adblock Plus " + minVersion + "]");
    else if(minVersion != "0")
      data.splice(0, 0, "(Adblock Plus " + minVersion + " or higher required) [Adblock]");
    else
      data.splice(0, 0, "[Adblock]");

    return data;
  },
  serializeArray: function(data, replaceVars) {
    if(replaceVars) {
      let now = new Date().toLocaleString();
      if(this.syncEnabled)
        now = "\$ " + now + " \$DATE";
      now = "$1" + now + "$2";
      data = data.map(function(pattern) {
        return pattern.replace(/^(!.*?\s)\$DATE(\s|$)/g, now);
      });
    }

    let abp = this.service.abp;
    return data.join(abp.Utils.getLineBreak());
  },
  getHash: function(data) {
    let abp = this.service.abp;
    return abp.Utils.generateChecksum(data.split(/[\r\n]+/g));
  },
  dataChanged: function(hash, data) {
    const PREF = "extensions.adblockfilters.lastUploadHash";
    if(!hash)
      hash = this.getHash(this.serializeArray(data || this.getData(), false));
    return (this.prefs.copyUnicharPref(PREF) != hash);
  },


  merge: function(remoteData, notModified, job) {
    let abp = this.service.abp;
    const VERSION_PREF = "extensions.adblockfilters.sync.version";
    let abort = false;

    let diff = function(filters1, filters2) {
      return filters1.filter(function(filter, index) {
        return filters2.indexOf(filter) == -1;
      });
    };

    LOG("parsing upload data...");
    let localFilters = this.parseFiltersFile(job.dataToUpload);
    let revision = this.prefs.getIntPref(VERSION_PREF);

    LOG("parsing data from last sync...");
    let baseFilters = this.parseFiltersFile(this.readFile(SYNC_BASE_FILE)).filters;
    baseFilters.pop(); // drop revision

    if(notModified) {
      abort = diff(localFilters.filters, baseFilters).length == 0 &&
              diff(baseFilters, localFilters.filters).length == 0;
    } else {
      LOG("parsing remote data...");
      let remoteFilters = this.parseFiltersFile(remoteData), remoteVersion;
      if(/^! Revision: (\d+)$/.test(remoteFilters.filters.pop())) {
        remoteVersion = parseInt(RegExp.$1);
      } else {
        throw Cr.NS_ERROR_UNEXPECTED;
      }
      let baseVersion = revision;
      revision = Math.max(revision, remoteVersion);

      if(baseVersion < remoteVersion) {
        let remoteRemovals = diff(baseFilters, remoteFilters.filters);
        let remoteAdditions = diff(remoteFilters.filters, baseFilters);

        if(!remoteAdditions.length && !remoteRemovals.length)
          abort = true;
        else {
          remoteRemovals.forEach(function(filter) {
            let(idx = localFilters.filters.indexOf(filter)) { if(idx != -1) localFilters.filters.splice(idx, 1) };
          });
          abp.AdblockPlus.removePatterns(remoteRemovals, remoteRemovals.length);

          remoteAdditions.forEach(function(filter) {
            if(localFilters.filters.indexOf(filter) == -1)
              localFilters.filters.push(filter);
          });
          abp.AdblockPlus.addPatterns(remoteAdditions, remoteAdditions.length);

          if(!this.useWhitelist) {
            this.blackWhiteList = this.blackWhiteList.filter(function(filter) {
              return (remoteAdditions.indexOf(filter) == -1);
            });
          } else {
            this.blackWhiteList = this.blackWhiteList.concat(remoteAdditions);
          }

          localFilters.filters.splice(0, 0, localFilters.header);
          job.init(localFilters.filters);
        }
      } else
        abort = diff(localFilters.filters, remoteFilters.filters).length == 0 &&
                diff(remoteFilters.filters, localFilters.filters).length == 0;
    }

    if(!abort)
      revision++;

    this.prefs.setIntPref(VERSION_PREF, revision);
    let lastLine = abp.Utils.getLineBreak() + "! Revision: " + revision;
    job._setData(job._dataNoChecksum + lastLine);

    if(abort)
      throw Cr.NS_ERROR_ABORT;
  },
  parseFiltersFile: function(data) {
    let abp = this.service.abp;
    let lines = data.split(/[\r\n]+/g);
    let obj = { filters: null, hash: "", header: "" };

    // regex stolen from abp:
    if(!/\[Adblock(?:\s*Plus\s*([\d\.]+)?)?\]/i.test(lines[0])) {
      LOG(lines[0]);
      throw Cr.NS_ERROR_FILE_UNKNOWN_TYPE;
    }

    lines = lines.filter(function(filter, index, lines) {
      return !isChecksum(filter, index, lines) && !!filter;
    });
    function isChecksum(filter, index, lines) {
      if(!/!\s*checksum[\s\-:]+([\w\+\/]+)/i.test(filter))
        return false;
      LOG(filter);
      let hash = obj.hash = RegExp.$1;
      lines = lines.slice(0); lines.splice(index, 1); // clone, then remove checksum
      if(hash && hash != abp.Utils.generateChecksum(lines)) {
        LOG("invalid checksum " + hash + " specified, " + abp.Utils.generateChecksum(lines) + " expected");
        throw Cr.NS_ERROR_FILE_CORRUPTED;
      }
      return !!hash;
    }
    lines = lines.map(function(filter) {
      return filter.replace(/^(!.*?\s)\$ (.+?) \$DATE(\s|$)/g, "$1\$DATE$3");
    });

    obj.filters = lines;
    obj.header = lines.shift(); // do this after generating all the checksums

    return obj;
  }
}