/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Malte Kraus.
 *
 * Contributor(s):
 *  Malte Kraus <firefox@maltekraus.de> (Original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
const Ci = Components.interfaces, Cc = Components.classes,
      Cr = Components.results, Cu = Components.utils;

function finish() {
  document.getElementById("upload-iframe").contentDocument.documentElement.acceptDialog();
  setTimeout(sync, 0, !!document.documentElement.getPageById("server-conf").next);
}
function doUpload(job) {
  const URL = "chrome://adblockfilters/content/ftp-upload.xul";
  window.openDialog(URL, URL, "chrome", job, false);
  nsPreferences.setIntPref("extensions.adblockfilters.sync.version", 1);
  window.close();
}
function sync(doSync) {
  let id = document.getElementById("sync-options").selectedItem.id;
  let action = id.substr(0, id.length - 6);
  let job = new Job(doSync, action);
  nsPreferences.setUnicharPref("extensions.adblockfilters.lastUploadHash", "");
  setTimeout(doUpload, 0, job);
}
function Job(doSync, action) {
  let data = service.getData();
  if(doSync) {
    data.push("! Revision: 1");
    this._revision = 1;
  }

  SyncJob.call(this, data);

  this.lastKnownRemote = 0;
  this.sync = doSync && (action != "local");

  this._action = action;
}
Job.prototype = {
  _action: "",
  _revision: 0,
  knownRemoteDate: 0,
  knownRemoteEtag: "",
  merge: function(newData) {
    let remote = service.parseFiltersFile(newData).filters;

    let foundRevision = false;
    if(remote.length && /^! Revision: (\d+)$/.test(remote[remote.length - 1])) {
      this._revision = parseInt(RegExp.$1);
      remote.pop(); // remove revision from array
      foundRevision = true;
    } else {
      LOG("Warning: remote file has no revision information... proceeding anyway");
    }

    let abp = Cc["@adblockfilters.mozdev.org/service;1"].getService().wrappedJSObject.abp;
    let bl = service.blackWhiteList;
    if(service.useWhitelist)
      bl = bl.concat(remote);
    else
      bl = bl.filter(function(f) remote.indexOf(f) == -1);
    service.blackWhiteList = bl;
    if(this._action == "remote") {
      let filter = bl;
      if(service.useWhitelist)
        filter = filter.filter(function(f) remote.indexOf(f) == -1);
      else
        filter = filter.concat(remote);

      abp.AdblockPlus.removePatterns(service.getData(null, filter));
      abp.AdblockPlus.addPatterns(remote);

      if(foundRevision) {
        service.blackWhiteList = bl;
        throw Cr.NS_ERROR_ABORT;
      }
    } else if(this._action == "merge") {
      abp.AdblockPlus.addPatterns(remote);

      if(foundRevision)
        this._revision = this._revision + 1;
    }
    service.blackWhiteList = bl;
    let data = service.getData(null, bl);
    data.push("! Revision: " + this._revision);
    this._setData(service.serializeArray(data, true));
  },
  observe: function(job, topic, errCode) {
    this.sync = true;
    SyncJob.prototype.observe.apply(this, arguments);
    if(topic != "adblockfilters-sync-finished" || job.wrappedJSObject != this ||
       parseInt(errCode) != Cr.NS_OK)
      return;
    const VERSION_PREF = "extensions.adblockfilters.sync.version";
    service.prefs.setIntPref(VERSION_PREF, this._revision);
    if(window && !window.closed)
      window.close();
  }
};

let service = Cc["@adblockfilters.mozdev.org/service;1"].getService().wrappedJSObject;
let servTop = ("getGlobalForObject" in Cu ? Cu.getGlobalForObject(service) : service.__parent__);
let LOG = servTop.LOG;
Job.prototype.__proto__ = new SyncJob();