/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Malte Kraus.
 *
 * Contributor(s):
 *  Malte Kraus <firefox@maltekraus.de> (Original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function savePassword() {
  var user = document.getElementById("textbox-ftp-user").value;
  var password = document.getElementById("textbox-ftp-pass").value;
  var host = document.getElementById("textbox-ftp-host").value;
  if("nsILoginManager" in Ci) {
    var passwordManager = Cc["@mozilla.org/login-manager;1"]
                            .getService(Ci.nsILoginManager);
    var pseudoURI = "chrome://adblockfilters";

    // remove old logins:
    var logins = passwordManager.findLogins({}, pseudoURI, pseudoURI, null);
    for(var i = 0; i < logins.length; i++) {
      passwordManager.removeLogin(logins[i]);
    }

    if(password) { // add new login:
      var loginInfo = Cc["@mozilla.org/login-manager/loginInfo;1"]
                        .createInstance(Ci.nsILoginInfo);
      loginInfo.init(pseudoURI, pseudoURI, null, user, password, "", "");
      passwordManager.addLogin(loginInfo);
    }
  } else {
    var tokendb = Cc["@mozilla.org/security/pk11tokendb;1"]
                    .createInstance(Ci.nsIPK11TokenDB);
    var token = tokendb.getInternalKeyToken();

    if(!token.isLoggedIn() && !token.checkPassword("")) {
      try {
        token.login(true);
      } catch(e) {
        if(!token.isLoggedIn())
          return;
      }
    }

    var passwordManager = Cc["@mozilla.org/passwordmanager;1"]
                            .createInstance(Ci.nsIPasswordManager);
    try { passwordManager.removeUser(host, user); } catch(e) {}
    if(password)
      passwordManager.addUser(host, user, password);
  }
}

function initPassword() {
  let user = document.getElementById("textbox-ftp-user").value;
  let textbox = document.getElementById("textbox-ftp-pass");
  let obj = Cc["@adblockfilters.mozdev.org/service;1"].getService().wrappedJSObject;
  let top = "getGlobalForObject" in Cu ? Cu.getGlobalForObject(obj) : obj.__parent__;
  let data = top.SyncJob.prototype.getUsernameAndPassword();
  if(data)
    textbox.value = textbox.defaultValue = data.password;
}

function syncStateChanged(enabled) {
  let check = document.getElementById("checkbox-sync-interval-enabled");
  check.disabled = !enabled;
  document.getElementById("textbox-sync-interval").disabled = !check.checked || !enabled;
}

function syncIntervalPref(read) {
  if(read)
    return document.getElementById("pref-sync-interval").value / 3600;
  else
    return document.getElementById("textbox-sync-interval").value * 3600;
}

function fileProtocol(protocol) {
  let file = (protocol == "file");
  document.getElementById("button-path-browse").hidden = !file;
  if(file) {
    document.getElementById("pref-ftp-host").value = "localhost";
    document.getElementById("fileProtocolActive").setAttribute("disabled", "true");
  } else
    document.getElementById("fileProtocolActive").removeAttribute("disabled");
}
function browseFile() {
  let fp = Cc["@mozilla.org/filepicker;1"].createInstance(Ci.nsIFilePicker);
  fp.init(window, "", Ci.nsIFilePicker.modeSave);
  fp.appendFilters(Ci.nsIFilePicker.filterAll | Ci.nsIFilePicker.filterText);
  fp.filterIndex = 1; // preselect text

  let url = "file:///" + document.getElementById("pref-ftp-path").value;
  try {
    let file = makeURI(url).QueryInterface(Ci.nsIFileURL).file;
    if(file.parent.exists()) fp.displayDirectory = file.parent;
    if(file.exists()) fp.defaultString = file.leafName;
  } catch(e) { }

  let rv = fp.show();
  if(rv != Ci.nsIFilePicker.returnOK)
    return;

  document.getElementById("pref-ftp-path").value = fp.fileURL.path.replace(/^\/+/, "/");
}

function initUploadPane() {
  initPassword();
  fileProtocol(document.getElementById("pref-ftp-protocol").value);
  syncStateChanged(document.getElementById("pref-sync-enabled").value);
  window.addEventListener("dialogaccept", savePassword, false);
}
function makeURI(uri, originCharset, baseURI) {
  let ioservice = Cc["@mozilla.org/network/io-service;1"]
                   .getService(Ci.nsIIOService);
  return ioservice.newURI(uri, originCharset || null, baseURI || null);
}

function showSyncWizard() {
  document.documentElement.currentPane.writePreferences(false);
  document.documentElement.openWindow("AdblockFilters:SyncSetup", "first-sync.xul",
                                      "first-sync.xul", "chrome");
}