/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Torisugari.
 *
 * Contributor(s):
 *  Torisugari (Original author)
 *  Nickolay Ponomarev <asqueella@gmail.com>
 *  Malte Kraus <firefox@maltekraus.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
 * Based on Infolister code which is based on BookmarksFTP code.
 */

/*
 * Provides data and target URI to upload to.
 */
function SyncJob() {
  this.wrappedJSObject = this;

  /*// only allow one job at a time
  if(SyncJob.current && !SyncJob.current.finished)
    return SyncJob.current;
  SyncJob.current = this;*/

  Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService)
            .addObserver(this, "adblockfilters-sync-finished", false);

  if(arguments.length)
    this.init.apply(this, arguments);
}
SyncJob.prototype = {
  classID: Components.ID("{3e0cb45d-4e51-fdf2-839b-5158df0c4d15}"),
  contractID: "@adblockfilters.mozdev.org/uploadjob;1",
  classDescription: "Adblock Plus Filter Uploader upload job",

  wrappedJSObject: null,

  init: function(data) {
    this.inProgress = true;

    this.sync = nsPreferences.getBoolPref("extensions.adblockfilters.sync.enabled");
    this.compress = nsPreferences.getBoolPref("extensions.adblockfilters.comp.enabled");
    this._setSourceAndTargetURI();

    this._setData(Cc["@adblockfilters.mozdev.org/service;1"].getService().wrappedJSObject.serializeArray(data, true));
  },
  /* used internally */
  _setSourceAndTargetURI: function() {
    const PREF_IX = "extensions.adblockfilters.ftp.";
    var protocol = nsPreferences.copyUnicharPref(PREF_IX + "protocol");
    var host = nsPreferences.copyUnicharPref(PREF_IX + "host");
    var path = nsPreferences.copyUnicharPref(PREF_IX + "path");

    if(!host || !path) throw Cr.NS_ERROR_NOT_AVAILABLE;

    var username = "";
    if(protocol != "file") {
      username = nsPreferences.copyUnicharPref(PREF_IX + "user");
      username = username ? encodeURIComponent(username) + "@" : "";
    }

    if(path[0] != "/") path = "/" + path; // fix path names missing the first /

    this.targetURI = makeURI(protocol + "://" + username + host + path);

    var url = nsPreferences.copyUnicharPref("extensions.adblockfilters.sync.download.url");
    this.sourceURI = url ? makeURI(url) : this.targetURI;
  },
  _dataNoChecksum: null,
  _setData: function (data) {
    if(data == this.dataToUpload || data == this._dataNoChecksum)
      return;

    var service = Cc["@adblockfilters.mozdev.org/service;1"].getService().wrappedJSObject;
    var nl = service.abp.Utils.getLineBreak();
    // strip previous checksum(s)
    data = data.replace(new RegExp(nl + "\\s*!\\s+Checksum:.+?" + nl, "g"), nl);
    this._dataNoChecksum = data;
    // compute checksum and update fields:
    this.writeHash = service.getHash(data);
    this.dataChanged = service.dataChanged(this.writeHash);
    // add checksum to data:
    this.dataToUpload = data.replace(nl, nl + "! Checksum: " + this.writeHash + nl);
  },

  /* properties for usage by sync.js */
  sync: false,
  compress: false,
  finished: false,
  dataChanged: true,
  writeHash: null,
  dataToUpload: "",
  sourceURI: null,
  targetURI: null,
  contentType: "text/plain",

  /* callbacks called by sync.js */
  merge: function(data, notModified) {
    adblockFilterServiceInstance.merge(data, notModified, this);
  },
  observe: function observe(job, topic, errCode) {
    if(topic != "adblockfilters-sync-finished" || job.wrappedJSObject != this)
      return;
    Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService)
                                         .removeObserver(this, topic, false);
    this.finished = true;
    if(parseInt(errCode) == Cr.NS_OK) {
      const PREF = "extensions.adblockfilters.";
      var changed = nsPreferences.copyUnicharPref(PREF + "lastUploadHash") != this.writeHash;
      if(changed)
        nsPreferences.setUnicharPref(PREF + "lastUploadHash", this.writeHash);
      if(this.sync) {
        var now = Math.round(new Date().getTime() / 1000);
        adblockFilterServiceInstance.lastSync = now;
        if(changed) {
          var fileName = ("getGlobalForObject" in Cu ? Cu.getGlobalForObject(adblockFilterServiceInstance) :
                          adblockFilterServiceInstance.__parent__).SYNC_BASE_FILE;
          adblockFilterServiceInstance.writeFile(fileName, this.dataToUpload);
        }
      } else { // synchronization needs to be re-initilatized when it's re-enabled
        nsPreferences.setIntPref("extensions.adblockfilters.sync.version", 0);
        adblockFilterServiceInstance.lastSync = 0;
      }
    }
  },
  getUsernameAndPassword: function() {
    const PREF_IX = "extensions.adblockfilters.ftp.";
    let username = null, password = null;
    try {
      if("nsILoginManager" in Ci) {
        let passwordManager = Cc["@mozilla.org/login-manager;1"]
                                .getService(Ci.nsILoginManager);
        // Find users for the given parameters
        let pseudoURI = "chrome://adblockfilters";
        let logins = passwordManager.findLogins({}, pseudoURI, pseudoURI, null);
        if(logins.length)
          return {user: logins[0].username, password: logins[0].password};
      } else if("@mozilla.org/passwordmanager;1" in Cc) {
        let token = Cc["@mozilla.org/security/pk11tokendb;1"]
                      .createInstance(Ci.nsIPK11TokenDB).getInternalKeyToken();

        if(!token.isLoggedIn() && !token.checkPassword(""))
          token.login(true);

        username = nsPreferences.copyUnicharPref(PREF_IX + "user");
        let host = nsPreferences.copyUnicharPref(PREF_IX + "host");
        let passwordManager = Cc["@mozilla.org/passwordmanager;1"]
                                .createInstance(Ci.nsIPasswordManager);
        let passwords = passwordManager.enumerator;
        while(password === null && passwords.hasMoreElements()) {
          let pass = passwords.getNext().QueryInterface(Ci.nsIPassword);
          if(pass.host == host && pass.user == username)
            return {user: username, password: pass.password};
        }
      }
    } catch(e) { Components.reportError(e); }
    return null; // no master password or no password found
  },

  get knownRemoteEtag() {
    return nsPreferences.copyUnicharPref("extensions.adblockfilters.sync.http.etag");
  },
  set knownRemoteEtag(val) {
    nsPreferences.setUnicharPref("extensions.adblockfilters.sync.http.etag", val);
    return val;
  },
  get knownRemoteDate() {
    return nsPreferences.copyUnicharPref("extensions.adblockfilters.sync.http.lastRemoteModification");
  },
  set knownRemoteDate(val) {
    nsPreferences.setUnicharPref("extensions.adblockfilters.sync.http.lastRemoteModification", val);
    return val;
  },

  notificationCallbacks: null,

  // nsISupports
  QueryInterface: function(aIID) {
    if(aIID.equals(Components.interfaces.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE;
  }
};