/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Torisugari.
 *
 * Contributor(s):
 *  Torisugari (Original author)
 *  Nickolay Ponomarev <asqueella@gmail.com>
 *  Malte Kraus <firefox@maltekraus.de>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
 * Based on Infolister upload.js which is based on Torisugari's BookmarksFTP.
 */

const Ci = Components.interfaces, Cc = Components.classes, Cr = Components.results;


function SyncService(aJob) {
  this.wrappedJSObject = this;
  this.job = aJob;
  this.notificationCallbacks.wrappedJSObject = this.notificationCallbacks;
  this.notificationCallbacks.sync = this;

  if(aJob.sync)
    this.retrieveData();
  else
    this.sendData();
}
SyncService.prototype = {
  wrappedJSObject: null,
  job: null,

  flagedExit: function(aCode) {
    Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService)
              .notifyObservers(this.job, "adblockfilters-sync-finished", aCode);
  },
  uriToText: function(uri) {
    var urispec = null;
    if(uri instanceof Ci.nsIURI) {
      uri.userPass = "";
      var textToUri = Cc["@mozilla.org/intl/texttosuburi;1"].getService(Ci.nsITextToSubURI);
      urispec = textToUri.unEscapeURIForUI(uri.originCharset, uri.spec);
    }
    return urispec;
  },

  dataToStream: function(data, binary) {
    if(binary) {
      let storageStream = Cc["@mozilla.org/storagestream;1"]
                            .createInstance(Ci.nsIStorageStream);
      storageStream.init(8192, -1, null);

      let os = Cc["@mozilla.org/binaryoutputstream;1"]
                 .createInstance(Ci.nsIBinaryOutputStream);
      os.setOutputStream(storageStream.getOutputStream(0));
      os[data instanceof Array ? "writeByteArray" : "writeBytes"](data, data.length);
      os.close();

      return storageStream.newInputStream(0);
    } else {
      let converter = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
                        .createInstance(Ci.nsIScriptableUnicodeConverter);
      converter.charset = "UTF-8";
      return converter.convertToInputStream(data);
    }
  },
  _convertStream: function(stream, from, to) {
    let output = "";
    let listener = {
      onStartRequest: function(request, context) { },
      onDataAvailable: function(request, context, stream, offset, count) {
        output += stream.QueryInterface(Ci.nsISupportsCString).data.substr(0, count);
      },
      onStopRequest: function(request, context, statusCode) { }
    };
    let streamConverter = Cc["@mozilla.org/streamConverters;1"]
                            .getService(Ci.nsIStreamConverterService);
    let compressor = streamConverter.asyncConvertData(from, to, listener, null);
    compressor.onStartRequest(null, null);
    compressor.onDataAvailable(null, null, stream, 0, stream.available());
    compressor.onStopRequest(null, null, Cr.NS_OK);
    return output;
  },
  compressStream: function(dataStream) {
    if(!("@mozilla.org/streamconv;1?from=uncompressed&to=rawdeflate" in Cc))
      return dataStream;

    let deflate = this._convertStream(dataStream, "uncompressed", "rawdeflate");

    // seek to the start so we can read from the same stream again
    dataStream.QueryInterface(Ci.nsISeekableStream);
    dataStream.seek(Ci.nsISeekableStream.NS_SEEK_SET, 0);

    // get raw UTF-8 bytes from the stream
    let bis = Cc["@mozilla.org/binaryinputstream;1"]
                .createInstance(Ci.nsIBinaryInputStream);
    bis.setInputStream(dataStream);
    let data = bis.readBytes(bis.available());
    bis.close();

    // gzip is deflate with some additional header data, see rfc 1952
    let crc = {};
    loadScript("crc32.js", crc);
    const GZIP_HEADER = "\x1f\x8b\x08\x01\0\0\0\0\0\0";
    let checksum = crc.numToByteStr(crc.crc32(data), 4);
    let size = crc.numToByteStr(data.length, 4);

    return this.dataToStream(GZIP_HEADER + deflate + checksum + size, true);
  },
  // uncompresses gzip data or returns input as it is
  uncompressByteArray: function(data) {
    let byteToStr = function(charCode) String.fromCharCode(charCode);
    if(data.slice(0, 3).map(byteToStr).join("") == "\x1f\x8b\x08") {
      let dataStream = this.dataToStream(data, true);
      data = this._convertStream(dataStream, "gzip", "uncompressed");
      dataStream.close();
    }
    return data;
  },
  _UTF8ToString: function(data) {
    let converter = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
                      .createInstance(Ci.nsIScriptableUnicodeConverter);
    converter.charset = "UTF-8";
    if(data instanceof Array)
      return converter.convertFromByteArray(data, data.length);
    else
      return converter.ConvertToUnicode(data);
  },

  newChannel: function(uri) {
    var iosvc = Cc["@mozilla.org/network/io-service;1"]
                  .getService(Ci.nsIIOService);
    var channel = iosvc.newChannelFromURI(uri);

    channel.loadFlags |= Ci.nsIRequest.VALIDATE_ALWAYS;
    channel.notificationCallbacks = this.notificationCallbacks;

    /*var authenticationManager = Cc["@mozilla.org/network/http-auth-manager;1"]
                                  .getService(Ci.nsIHttpAuthManager);
    authenticationManager.clearAll();*/

    this.channel = channel;
    return channel;
  },

  STEP_SEND: 2,
  // uploads contents of job.streamToUpload to job.targetURI
  sendData: function() {
    this.streamListener.step = this.STEP_SEND;

    var uri = this.job.targetURI;
    var contentType = this.job.contentType;

    var stream = this.dataToStream(this.job.dataToUpload, false);
    if(this.job.compress) {
      stream = this.compressStream(stream);
      contentType = "application/gzip";
    }

    if(!stream || !uri) {
      LOG("no data or uri");
      this.flagedExit(-1);
      return;
    }

    // some of our streams have a toString method that throws, so use the native one:
    LOG("sendData(): uploading " + uri.toString.call(stream) + " to " + this.uriToText(uri) + ".");
    try {
      var channel = this.newChannel(uri);
      channel.QueryInterface(Ci.nsIUploadChannel);

      // prevent race conditions for http, can't do anything for other protocols:
      if(this.job.sync && channel instanceof Ci.nsIHttpChannel) {
        if(this.job.knownRemoteDate)
          channel.setRequestHeader("If-Unmodified-Since", this.job.knownRemoteDate, true);
        if(this.job.knownRemoteEtag)
          channel.setRequestHeader("If-Match", this.job.knownRemoteEtag, true);
      }

      LOG("sendData(): calling " + channel + ".setUploadStream()");
      channel.setUploadStream(stream, contentType, -1);

      LOG("calling asyncOpen()...");
      setStatusText("send", null);
      channel.asyncOpen(this.streamListener, this);
    } catch(e) {
      LOG("uri=" + this.uriToText(uri));
      Components.reportError(e);
      this.flagedExit(-1);
    }
  },

  STEP_RETRIEVE: 1,
  retrieveData: function() {
    this.streamListener.step = this.STEP_RETRIEVE;
    var uri = this.job.sourceURI;
    LOG("retrieveData(): downloading data from " + this.uriToText(uri) + ".");

    if(!uri) {
      LOG("no uri");
      this.flagedExit(-1);
      return;
    }

    try {
      var channel = this.newChannel(uri);
      if(channel instanceof Ci.nsIHttpChannel) {
        if(this.job.knownRemoteDate)
          channel.setRequestHeader("If-Modified-Since", this.job.knownRemoteDate, true);
        if(this.job.knownRemoteEtag)
          channel.setRequestHeader("If-Non-Match", this.job.knownRemoteEtag, true);
      }

      LOG("calling asyncOpen()...");
      setStatusText("load", null);
      channel.asyncOpen(this.streamListener, this);
    } catch(e) {
      LOG("uri=" + this.uriToText(uri));
      Components.reportError(e);
      this.flagedExit(-2);
    }
  },
  dataRetrieved: function(data) {
    LOG("dataRetrieved() called");
    setStatusText("merge", null);
    let abort = false, error = false;
    let notModified = false;
    if(this.channel instanceof Ci.nsIHttpChannel) {
      notModified = (this.channel.responseStatus == 304);
    }
    try {
      this.job.merge(data, notModified);
    } catch(e if e == Cr.NS_ERROR_ABORT) {
      abort = true;
      LOG("remote and local data are identical");
    } catch(e) {
      error = true;
      LOG("error with retrieved data: " + e);
    }

    if(!error) {
      let channel = this.channel;
      let getHeader = function(header) {
        try {
          return channel.getResponseHeader(header);
        } catch(e if e.code == Ci.NS_ERROR_NOT_AVAILABLE) { }
        return "";
      };
      this.job.knownRemoteDate = getHeader("Last-Modified") || getHeader("Date");
      this.job.knownRemoteEtag = getHeader("Etag");
    }
    if(abort || error)
      this.flagedExit(error ? -3 : Cr.NS_OK);
    else
      this.sendData();
  },

  cancel: function() {
    if(this.channel)
      this.channel.cancel(Cr.NS_BINDING_ABORTED);
  },

  streamListener: {
    step: -1,
    bis: Cc["@mozilla.org/binaryinputstream;1"].createInstance(Ci.nsIBinaryInputStream),
    data: null,
    // nsIStreamListener
    onDataAvailable: function(aChannel, aCtxt, aInStr, aSourceOffset, aCount) {
      if(this.step == SyncService.prototype.STEP_RETRIEVE) {
        this.bis.setInputStream(aInStr);
        let bytes = this.bis.readByteArray(aCount);
        this.data = this.data ? this.data.concat(bytes) : bytes;
      }
    },
    hasSucceeded: function(statusCode) {
      return (statusCode == Cr.NS_OK ||
              statusCode == Cr.NS_ERROR_INVALID_CONTENT_ENCODING);
    },
    // nsIRequestObserver
    onStartRequest: function(channel, sync) { },
    onStopRequest: function(aChannel, aCtxt, aErrCode) {
      LOG("onStopRequest(" + aErrCode + ") step: " + this.step);
      aCtxt = aCtxt.wrappedJSObject;

      try {
        if(aChannel instanceof Ci.nsIHttpChannel && !aChannel.requestSucceeded &&
           aChannel.responseStatus != 304 && aChannel.responseStatus != 301) {
          LOG("http error: " + aChannel.responseStatus + " " + aChannel.responseStatusText);
          aCtxt.flagedExit(aChannel.responseStatus);
          return;
        }
      } catch(e if e.code == Ci.NS_ERROR_NOT_AVAILABLE) { }

      if(!this.hasSucceeded(aErrCode)) {
        LOG("error retrieving data: " + aErrCode);
        aCtxt.flagedExit(aErrCode);
        return;
      }

      if(this.step == aCtxt.STEP_RETRIEVE) {
        // try uncompressing the data, so that syncing "just works" even for a
        // misconfigured server (subscribing to it with ABP will still fail)
        let data = aCtxt.uncompressByteArray(this.data || []);
        data = aCtxt._UTF8ToString(data);
        aCtxt.dataRetrieved(data);
        delete this.data;
      } else {
        let job = aCtxt.wrappedJSObject.job;
        job.knownRemoteEtag = job.knownRemoteDate = "";
        if(aChannel instanceof Ci.nsIHttpChannel) {
          try {
            job.knownRemoteDate = aChannel.getResponseHeader("Date");
          } catch(e if e.code == Ci.NS_ERROR_NOT_AVAILABLE) { }
        }

        aCtxt.flagedExit(Cr.NS_OK);
      }
    }
  },
  notificationCallbacks: {
    sync: null, // parent object
    // nsIInterfaceRequestor
    getInterface: function(aIID) {
      try {
        if(this.sync.job.notificationCallbacks)
          return this.sync.job.notificationCallbacks.QueryInterface(aIID);
      } catch(e) {};
      try {
        return this.QueryInterface(aIID)
      } catch(e) {}
      return null;
    },
    // nsIAuthPrompt
    prompt: function() { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
    promptPassword: function() { throw Cr.NS_ERROR_NOT_IMPLEMENTED; },
    _retries: 0,
    promptUsernameAndPassword: function(title, text, realm, savePw, user, password) {
      var login = this.sync.job.getUsernameAndPassword();
      if(!login) {
        let authprompt = Cc["@mozilla.org/embedcomp/window-watcher;1"]
                           .getService(Ci.nsIWindowWatcher)
                           .getNewAuthPrompter(null);
        return authprompt.promptUsernameAndPassword(title, text, realm, savePw,
                                                    user, password);
      }
      if(++this._retries >= 3)
        throw Cr.NS_BINDING_ABORTED;
      user.value = login.user;
      password.value = login.password;
      return true;
    },
    //nsIAuthPrompt2
    asyncPromptAuth: function(channel, callback, context, level, authInfo) {
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    },
    promptAuth: function(channel, level, authInfo) {
      if(++this._retries >= 3)
        return false;

      var login = this.sync.job.getUsernameAndPassword();
      if(!login) {
        let authprompt = Cc["@mozilla.org/embedcomp/window-watcher;1"]
                           .getService(Ci.nsIWindowWatcher)
                           .getNewAuthPrompter(null)
                           .QueryInterface(Ci.nsIAuthPrompt2);
        return authprompt.promptAuth(channel, level, authInfo);
      }
      authInfo.username = login.user;
      authInfo.password = login.password;

      return true;
    },

    // nsISupports
    QueryInterface: function(aIID) {
      if(aIID.equals(Ci.nsIAuthPrompt) || aIID.equals(Ci.nsIAuthPrompt2) ||
         aIID.equals(Ci.nsISupports) || aIID.equals(Ci.nsIInterfaceRequestor))
        return this;
      throw Cr.NS_ERROR_NO_INTERFACE;
    }
  }
};


/* depends on these functions to be available:
   (use nsIScriptLoader if you don't want to pollute your namespace)
  - function setStatusText //
  - function LOG // debugging, use empty function body for production
  - function loadScript // wrapper for mozIJSSubScriptLoader
*/